


/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                     University of Edinburgh, UK                       */
/*                        Copyright (c) 1999                             */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission is hereby granted, free of charge, to use and distribute  */
/*  this software and its documentation without restriction, including   */
/*  without limitation the rights to use, copy, modify, merge, publish,  */
/*  distribute, sublicense, and/or sell copies of this work, and to      */
/*  permit persons to whom this work is furnished to do so, subject to   */
/*  the following conditions:                                            */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*   4. The authors' names are not used to endorse or promote products   */
/*      derived from this software without specific prior written        */
/*      permission.                                                      */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*             Author :  Alan W Black (awb@cstr.ed.ac.uk)                */
/*             Date   :  March 1999                                      */
/*-----------------------------------------------------------------------*/
/*                                                                       */
/* Client end of Festival server API (in C) designed specifically for    */
/* Galaxy Communicator use, though might be of use for other things      */
/*                                                                       */
/*=======================================================================*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>



#ifndef _FESTIVAL_CLIENT_H_
#define _FESTIVAL_CLIENT_H_

#define FESTIVAL_DEFAULT_SERVER_HOST "localhost"
#define FESTIVAL_DEFAULT_SERVER_PORT 1314
//#define FESTIVAL_DEFAULT_TEXT_MODE "fundamental"
#define FESTIVAL_DEFAULT_TEXT_MODE "async"


typedef struct FT_Info
{
    int encoding;
    char *server_host;
    int server_port;
    char *text_mode;
    
    int server_fd;
} FT_Info;

typedef struct FT_Wave
{
    int num_samples;
    int sample_rate;
    short *samples;
} FT_Wave;

void delete_FT_Wave(FT_Wave *wave);
void delete_FT_Info(FT_Info *info);

#define SWAPSHORT(x) ((((unsigned)x) & 0xff) << 8 | \
                      (((unsigned)x) & 0xff00) >> 8)
#define SWAPINT(x) ((((unsigned)x) & 0xff) << 24 | \
                    (((unsigned)x) & 0xff00) << 8 | \
		    (((unsigned)x) & 0xff0000) >> 8 | \
                    (((unsigned)x) & 0xff000000) >> 24)

/* Sun, HP, SGI Mips, M68000 */
#define FAPI_BIG_ENDIAN (((char *)&fapi_endian_loc)[0] == 0)
/* Intel, Alpha, DEC Mips, Vax */
#define FAPI_LITTLE_ENDIAN (((char *)&fapi_endian_loc)[0] != 0)


/*****************************************************************/
/*  Public functions to interface                                */
/*****************************************************************/

/* If called with NULL will attempt to access using defaults */
FT_Info *festivalOpen(FT_Info *info);
FT_Wave *festivalStringToWave(FT_Info *info,char *text);
int festivalClose(FT_Info *info);

#endif






static FT_Info *festival_default_info();
static int festival_socket_open(const char *, int);
FT_Info *festivalOpen(FT_Info *);
int festivalSpeech(FT_Info *,char *);
int festivalCommand(FT_Info *,char *);
int festivalClose(FT_Info *);


FT_Info *info;



static FT_Info *festival_default_info()
{
    FT_Info *info;
    info = (FT_Info *)malloc(1 * sizeof(FT_Info));
    
    info->server_host = FESTIVAL_DEFAULT_SERVER_HOST;
    info->server_port = FESTIVAL_DEFAULT_SERVER_PORT;
    info->text_mode = FESTIVAL_DEFAULT_TEXT_MODE;

    info->server_fd = -1;
    
    return info;
}

static int festival_socket_open(const char *host, int port)
{   
    /* Return an FD to a remote server */
    struct sockaddr_in serv_addr;
    struct hostent *serverhost;
    int fd;

    fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    if (fd < 0)  
    {
	fprintf(stderr,"festival_client: can't get socket\n");
	return -1;
    }
    memset(&serv_addr, 0, sizeof(serv_addr));
    if ((serv_addr.sin_addr.s_addr = inet_addr(host)) == -1)
    {
	/* its a name rather than an ipnum */
	serverhost = gethostbyname(host);
	if (serverhost == (struct hostent *)0)
	{
	    fprintf(stderr,"festival_client: gethostbyname failed\n");
	    return -1;
	}
	memmove(&serv_addr.sin_addr,serverhost->h_addr, serverhost->h_length);
    }
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(port);

    if (connect(fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) != 0)
    {
	fprintf(stderr,"festival_client: connect to server failed\n");
	return -1;
    }

    return fd;
}




/***********************************************************************/
/* Public Functions to this API                                        */
/***********************************************************************/

FT_Info *festivalOpen(FT_Info *info)
{
    /* Open socket to server */

    if (info == 0)
	info = festival_default_info();

    info->server_fd = 
	festival_socket_open(info->server_host, info->server_port);
    if (info->server_fd == -1)
	return NULL;

    return info;
}



int festivalSpeech(FT_Info *info,char *text) {

    FT_Wave *wave;
    FILE *fd;
    char ack[4];
    char *p;
    int n;


    if (info == 0)
	return -1;

    if (info->server_fd == -1) {
	fprintf(stderr,"festival_client: server connection unopened\n");
	return -1;
    }
    
    fd = fdopen(dup(info->server_fd),"wb");


    /* Copy text over to server, escaping any quotes */
    
    fprintf(fd,"(SayText \"");  
    for (p=text; p && (*p != '\0'); p++)
    {
	if ((*p == '"') || (*p == '\\'))
	    putc('\\',fd);

	putc(*p,fd);
    }
    fprintf(fd,"\")\n");   
    fclose(fd);

    
    return 0;
}



int festivalCommand(FT_Info *info,char *text) {

    FT_Wave *wave;
    FILE *fd;
    char ack[4];
    char *p;
    int n;


    if (info == 0)
	return -1;

    if (info->server_fd == -1) {
	fprintf(stderr,"festival_client: server connection unopened\n");
	return -1;
    }
    
    fd = fdopen(dup(info->server_fd),"wb");


    /* Copy text over to server, escaping any quotes */

    fprintf(fd,text);  
//    printf("festivalCommand: %s\n", text);
    
/*    fprintf(fd,"(");  
    for (p=text; p && (*p != '\0'); p++)
    {
	if ((*p == '"') || (*p == '\\'))
	    putc('\\',fd);
	putc(*p,fd);
    }
    fprintf(fd,")\n");   
*/    fclose(fd);

    
    return 0;
}




int festivalClose(FT_Info *info)
{
    if (info == 0)
	return 0;

    if (info->server_fd != -1)
	close(info->server_fd);

    return 0;
}


