
/*

	Middleware 0.95
	Autore: Matteo Fasanella
	E-Mail: mfasanella@users.sourceforge.net
	Licenza: GNU/GPL

*/



#include <signal.h>
#include <fcntl.h>

#include <dlfcn.h>

#include <sys/sysmacros.h>

//#include <sys/time.h>
#include <unistd.h>

#include "festival.h"
#include "parser.h"
#include <sys/linux4d.h>

#define spkupLogFile	"speakup.log"		// logfile x speakup (sigusr1)
#define midwrLogFile	"middleware.log"	// logfile x middleware (sigusr2)

#define firstCon	1
#define lastCon		49

#define maxModules	50

#define L4Ddebug	1

int init_all();
int close_all();
static void sig_usr1(int);
static void sig_usr2(int);
static void sig_hup(int);
static void sig_urg(int);
static void sig_pipe(int);
static void sig_chld(int);
void gestCons(int);
void gestFn(char *, int);
char *makeAcc(char *);
void load_modules(char *);
int find_pid (int, char *, char *); 
int check_EXT(int, int, int, char);
int readscreen(char *);
int speak(char *);
static void fire_speak(int);

int fd1, fd2;
int cons=0;
char *fmState="";
int log = 1;			// logga
char *systemVoice, *userVoice;
int Active = 1;

int numrows = 25, numcols = 80;

int vcsSR[25];
int vcsOffs[25];

char festival_buf[4096];
unsigned int fest_delay = 1;

int flag[50];			// per l'esecuzione condizionata delle azioni
struct domain *Fnbase[50];	// max 50 Console + applicaz. X
struct head *Fncurr[50];

int (*mod_fun[maxModules])();
char mod_names[maxModules][30];	// array di maxModules nomi lunghi 30
int *handle[maxModules];



int main (void) {
    
    int tmppid = 0;

    init_all();



    if (festivalCommand(info, systemVoice) != 0) printf("Errore nella Speech!\n");    		    	
    if (festivalSpeech(info, "Inizializzazione completata..... ") != 0) printf("Errore nella Speech!\n");    		    


    for(;;) pause();
    
    close_all();

return 0;
}






int init_all() {

    char filetmp[32]="", *conname = (char *) malloc(20);
    int i=0;



    if ( freopen("/dev/tty24", "r+", stdout)  < 0 ) { perror("Non e` possibile aprire STDOUT"); exit(-1); }
    if ( freopen("/dev/tty24", "r+", stderr)  < 0 ) { perror("Non e` possibile aprire STDOUT"); exit(-1); }    

    // init signal handlers
    if (signal(SIGUSR1, sig_usr1) == SIG_ERR) { printf("can't catch SIGUSR1\n"); exit(-1); }
    if (signal(SIGUSR2, sig_usr2) == SIG_ERR) { printf("can't catch SIGUSR2\n"); exit(-1); }
    if (signal(SIGHUP,  sig_hup)  == SIG_ERR) { printf("can't catch SIGHUP \n"); exit(-1); }
    if (signal(SIGURG,  sig_urg)  == SIG_ERR) { printf("can't catch SIGURG \n"); exit(-1); }    
    if (signal(SIGPIPE,  sig_pipe) == SIG_ERR) { printf("can't catch SIGPIPE \n"); exit(-1); }
    if (signal(SIGCHLD,  sig_chld) == SIG_ERR) { printf("can't catch SIGCHLD \n"); exit(-1); }    
//    if (signal(SIGALRM,  fire_speak) == SIG_ERR) { printf("can't catch SIGALRM \n"); exit(-1); }    




    //  init log files
//    if ( (fd1=open(spkupLogFile, O_RDWR|O_CREAT, 00666) ) < 0 ) perror("No text log file");
    if ( (fd2=open(midwrLogFile, O_RDWR|O_CREAT, 00666) ) < 0 ) perror("No ctrl log file");


    // init festival 
    info = festival_default_info();
    printf("server: %s, on port: %d, with mode: %s\n", info->server_host, info->server_port, info->text_mode);
//    if ( (info = festivalOpen(info)) == NULL) { printf("Festival Daemon not loaded\n"); exit(-1); }
    if ( (info = festivalOpen(info)) == NULL)  printf("Festival Daemon not loaded\n"); 
    if (festivalCommand(info, "(audio_mode 'async)\n") != 0) printf("Errore nella Speech!\n");    		    	


    
    // init RuleFile Structure
    for (i=firstCon; i<=lastCon; i++) {
	bzero(filetmp, 32);
	sprintf(filetmp, "rules/f%d.rul", i);
        Fnbase[i] = create_rules(filetmp);
	if (Fnbase[i]->headptr !=  NULL) {
	    Fncurr[i] = Fnbase[i]->headptr;
	    printf("init console %d, with: %s => %s  <%s>\n", i, filetmp, Fnbase[i]->con_name, Fnbase[i]->app_name); 
//   		print_rules(Fnbase[i]);
	} 
    flag[i] = 1;		// attiva i flag
    }


    // load Modules
    load_modules("rules/modules");


    systemVoice = (char *) malloc(128); systemVoice = "(voice_lp_mbrola)\n";
    userVoice = (char *) malloc(128);	userVoice = "(voice_pc_mbrola)\n";


    // usare setitimer per una granularita` piu` fine
    fest_delay = 1;
//    alarm(fest_delay);
    



    // lasciare queste inizializzazioni per ultime,
    // altrimenti le righe successive non funzionano come dovrebbero
    for(i=1; i<25; i++) {
	vcsOffs[i] = 0;
	sprintf(conname, "/dev/vcs%d", i);
	if ( (vcsSR[i] = open(conname, O_RDONLY, 0)) < 0 ) { perror(""); printf("so I can't open /dev/vcs%d\n", i);  return 0;}
        lseek(vcsSR[i], 0, SEEK_SET);	
    }

return 0;
}


int close_all() {
    int i=-1;

    festivalClose(info);
    close(fd1);
    close(fd2);    

    for(i=1; i<25; i++) {
	close(vcsSR[i]);
    }
    
    while (handle[++i] != NULL )   dlclose(handle[i]);

return 0;
}



static void sig_pipe(int signo) {

    if (signo==SIGPIPE) { 
	printf("SIGPIPE Caught! Middleware can Continue...\n\n");    
	Active = 1;
    }

return;
}


static void sig_chld(int signo) {

    if (signo==SIGCHLD) { 
	printf("SIGCHLD Caught! Middleware Paused...\n\n");
	Active = 0;
    }

return;
}




static void sig_usr1(int signo) {

    char *ritorno=(char *) malloc(8192);

    if ( (signo==SIGUSR1) && (Active) ) {
//	get2speak(ritorno);   
        printf("get2speak: %s\n", ritorno);

	if (log) {
//	    write (fd1, ritorno, strlen(ritorno)); write (fd1, "\n", strlen("\n"));	
	}

	if (festivalSpeech(info, ritorno) != 0) printf("Errore nella Speech!\n");    	
    }
    
    return;
}

static void sig_hup(int signo) {

    if (signo==SIGHUP) {
        printf("sighup\n\n");
	
	load_modules("rules/modules");		// "dovrebbe" ricaricare tutti i moduli

	if (log) {
	    write (fd2, "SIGHUP", strlen("SIGHUP")); write (fd2, "\n", strlen("\n"));	
	}

	if (festivalSpeech(info, "Sto aggiornando il Middleware") != 0) printf("Errore nella Speech!\n");    	
    }
    
    return;
}


static void sig_urg(int signo) {

    char tmp, prog[128], tty[128], buf[128];
    int	pid=0, fd, r=0, count, k=1, ttyn=0, i=0;

    if ( (signo==SIGURG) && (Active) ) {
        printf("\n\nSIGURG Caught!\n\n");
	
	if (log) {
	    write (fd2, "SIGURG", strlen("SIGURG")); write (fd2, "\n", strlen("\n"));	
	}

        if((fd=open("/var/tmp/exec.pid", O_RDONLY)) < 0) { perror("errore nell'apertura di /var/tmp/exec.pid"); return;}
	r=read(fd, &tmp, 1);
	while ( (tmp>='0') && (tmp<='9') && (r>0) ) { 
	    pid = (pid*10) + (tmp-'0');
	    r=read(fd, &tmp, 1); 
	}
        close(fd);
	
	// andando a prendere il nome in "stat", funziona anche quando ci sono
	// problemi con la linea di comando
	sprintf(buf, "/proc/%d/stat", pid);
        if ( (fd = open(buf, O_RDONLY)) < 1) {perror("errore nella open"); return;}    
        read(fd, &prog[0], 1);
	while (prog[0] != '(') read(fd, &prog[0], 1);
	read(fd, &prog[0], 1);
	while (prog[k-1] != ')')  read(fd, &prog[k++], 1);
	prog[k-1] = '\0';
        close(fd);

        bzero(buf, 128);      sprintf(buf, "/proc/%d/fd/%d", pid, 2);
	if((count=readlink(buf, tty, 1024)) == -1) { perror("errore nella readlink"); return;}    
        tty[count] = '\0';

	printf("PID = %d, prog = %s, tty = %s\n\n", pid, prog, tty);
	
	
//	if (festivalSpeech(info, "Sto aggiornando il Middleware") != 0) printf("Errore nella Speech!\n");    	
    
        while( (tty[i]<'0') || (tty[i]>'9') ) i++;
	while( (tty[i]>='0') && (tty[i]<='9') ) ttyn = (ttyn*10) + (tty[i++] - '0');
	bzero(buf, 128);      sprintf(buf, "rules/%s.rul", prog);
        Fnbase[ttyn] = create_rules(buf);
	if (Fnbase[ttyn]->headptr !=  NULL) {
    	    Fncurr[ttyn] = Fnbase[ttyn]->headptr;
	    printf("init console %d, with: %s => %s  <%s>\n", ttyn, buf, Fnbase[ttyn]->con_name, Fnbase[ttyn]->app_name); 
//    	    print_rules(Fnbase[ttyn]);
	} else printf("non trovo il file di regole per %s\n\n", prog);

	return;
    }
    

    return;
}



static void sig_usr2(int signo) {

    char *ritorno=(char *) malloc(8192);
    

    if ( (signo==SIGUSR2) && (Active) ) {
	get2ctrl(ritorno);   
	printf("get2ctrl: %s\n", ritorno);
	
	if ( (ritorno[4] >= 1) && (ritorno[4] <= 27) ) {	// per i '^...'
	    printf("ritorno[4]: %d\n\n", ritorno[4]);
	    if ( ritorno[4] == 27)  sprintf(ritorno, "Self^[");	    
	    else if ( ritorno[4] == 9)  sprintf(ritorno, "K_TAB");	    
	    else  sprintf(ritorno, "Self^%c", 'A'+ritorno[4]-1);	    
	}
        if ( ritorno[4] == 127)  sprintf(ritorno, "Self^<");	    
	
	
	if (log) {
	    write (fd2, ritorno, strlen(ritorno)); write (fd2, "\n", strlen("\n"));	
	}

	// se c'e` cambio di console (=> voce di sistema)
	if (strncmp(ritorno, "Cons", 4) == 0) {
	    ((strlen(ritorno)==5)?(cons = (ritorno[4] - '0')):(cons = ((ritorno[4] - '0')*10)+(ritorno[5] - '0')) );
	    gestCons(cons);
	    return;
	}

	// se c'e` un comando per lo Screen Reader
	if ( (strncmp(ritorno, "Pad", 3) == 0) && ((ritorno[3] - '0') >= 0) && ((ritorno[3] - '0') <= 9) ) {
	    printf("readscreen(%s)\n", &ritorno[3]);
	    readscreen(&ritorno[3]);  // lo posso fare perche` ritorno[4]='\0'
	    return;
	}

	// se c'e` evento ad un servizio 
	if ( (cons >= firstCon) && (cons <= lastCon) ) {
	    gestFn(ritorno, cons);	// e lo passa a gestFn
	    return;
	}

	// altrimenti, se e` su una console sconosciuta legge il comando
	if (festivalSpeech(info, ritorno) != 0) printf("Errore nella Speech!\n");    	
    }
    
    return;
}



void gestCons(int consn) {

    char toRead[1024]="";

    if (!Active) return;

    sprintf(toRead, "terminale..... %d..", consn);

    if ( (consn >= firstCon) && (consn <= lastCon) ) {
	if (strcmp(Fnbase[consn]->con_name, "") == 0) strcpy(Fnbase[consn]->con_name, toRead);
	printf("%s\n\n", Fnbase[consn]->con_name);
        if (festivalCommand(info, systemVoice) != 0) printf("Errore nella Speech!\n");    		    	
//	if (festivalCommand(info, "(Parameter.set \"Duration_Stretch\" 1.2)\n") != 0) printf("Errore nella Speech!\n");    		    	

        if (festivalSpeech(info, makeAcc(Fnbase[consn]->con_name)) != 0) printf("Errore nella Speech!\n");    		    
	return;
    }
    
    printf("Console %d\n\n", consn);
    if (festivalSpeech(info, toRead) != 0) printf("Errore nella Speech!\n");    	

return;
}




void gestFn(char *event, int con) {

struct cell *currcell=NULL;
struct action *curract=NULL;
int esci=0, i=0, pid = 1000000;
char toRead[1024]="", tmp[128];

    if (!Active) return;

    if (festivalCommand(info, userVoice) != 0) printf("Errore nella Speech!\n");    		    	
    if (festivalCommand(info, "(Parameter.set \"Duration_Stretch\" 1.1)\n") != 0) printf("Errore nella Speech!\n");    		    	

    if ((Fnbase[con]->headptr) == NULL) { 
        if (festivalSpeech(info, "Nessun file di regole per questo terminale!") != 0) printf("Errore nella Speech!\n");    		    
	printf("Nessun file di regole per questo terminale!\n"); 
	return;
    }

    if (strcmp(event, "Self ") == 0) strcpy(event, "Self"); // per lo spazio

    currcell=Fncurr[con]->table;

    if (strcmp(event, "Fn11") == 0) { 	// Reset dello stato dell'app. corr.
	Fncurr[con] = Fnbase[con]->headptr;
	sprintf(tmp, "/dev/tty%d", con);
	printf("Reset dello stato di %s su %s\n\n", Fnbase[con]->app_name, tmp);
	if ( (pid = find_pid(0, Fnbase[con]->app_name, tmp)) > 0 ) 
	    kill (pid, 9);
	goto end_events;
    }

    

			// il controllo su "Char" e` da ottimizzare
    if ( (strncmp(event, "Self", 4) == 0) && ( ((event[4]>='a') && (event[4]<='z')) || (event[4]>='A' && event[4]<='Z') ) ) {	
        while ( (currcell != NULL) && (!esci) ) {
    	    if (strcmp(currcell->event, "Char") == 0) esci=1;
	    else    if (currcell->next_cell != NULL)
			currcell=currcell->next_cell;
		    else esci=1;
	}
	esci = 0;
	if (strcmp(currcell->event, "Char") == 0)  goto charjmp;
    } 


    currcell=Fncurr[con]->table;
    
    while ( (currcell != NULL) && (!esci) ) {
    	    if (strcmp(currcell->event, event) == 0) esci=1;
	    else    if (currcell->next_cell != NULL)
			currcell=currcell->next_cell;
		    else esci=1;
    }
    if (strcmp(currcell->event, event) != 0) {
        if (festivalSpeech(info, "Evento sconosciuto....") != 0) printf("Errore nella Speech!\n");    		    
        printf("Unknown event, state unchanged\n");
        goto end_events;
    }

    charjmp:

    curract = currcell->actions;
    while ( (curract != NULL) && (flag[con]) )  {
		
//	printf("event, action: %s, %s\n", currcell->event, curract->action);    

	if (strcmp(curract->action, "nil") == 0) goto end_events;
    
        if (strcmp(curract->action, "go") == 0) {	// evento "go"
    	    Fncurr[con]=curract->next_table;
	    printf("go to: [%s]\n\n", Fncurr[con]->header);
	    
	    goto end_events;

	} else {  
	
	    // controllo delle funzioni aggiunte dai Moduli
	    esci=0; i=-1;
	    while ( (!esci) && (++i<maxModules) ) {
		if (strcmp(mod_names[i], curract->action) == 0 ) 
		    esci = 1;
	    }
	    if (esci) {	// azione da modulo
		printf("%s(%s) = %d\n\n", mod_names[i], curract->parms, mod_fun[i](curract->parms));		
	    
	    
	    } else {	// azione sconosciuta
	        sprintf(toRead, "Funzione... sconosciuta... %s..... ", curract->action);
		printf("funzione sconosciuta: %s(%s)\n\n", curract->action, curract->parms);
    		if (festivalSpeech(info, toRead) != 0) printf("Errore nella Speech!\n");    		    	
	    }
	}
    
	curract = curract->next_action;
	
    }	// end while    

    end_events:    
    flag[con] = 1;
    
    esci = 0;	//  controllo se ci sono eventi EXT da servire
    currcell=Fncurr[con]->table;
    
    while ( (currcell != NULL) && (!esci) ) {
        if (strcmp(currcell->event, "EXT") == 0) {
	    esci=check_EXT(con, currcell->x, currcell->y, currcell->token);
	    if ( (!esci) || (currcell->next_cell != NULL) ) 
		    currcell=currcell->next_cell;
		else currcell=NULL;
	}
        else    if (currcell->next_cell != NULL)
		    currcell=currcell->next_cell;
		else currcell=NULL;
    }
    if ( (!esci) || (currcell == NULL) ) { 
	return;
    }
    else goto charjmp;
    

return;
}



char *makeAcc(char *toacc) {  // make a phrase, readable

    char *toret = (char *) malloc(4096);
    int i=0, toreti=0;

    while (i<strlen(toacc)) {

//	if (!(((toacc[i]>='a') && (toacc[i]<='z')) || ((toacc[i]>='A') && (toacc[i]<='Z')) || (toacc[i]=='.') || ((toacc[i]>='0') && (toacc[i]<='9')) ) )
//	    toacc[i] = ' ';
	if (toacc[i] == '/') {
	    toret[toreti++] = ' '; toret[toreti++] = 's'; toret[toreti++] = 'l'; 
	    toret[toreti++] = 'e'; toret[toreti++] = 's'; toret[toreti++] = 'c'; 
	    toret[toreti++] = 'i'; toret[toreti++] = ' ';
	    i++;

	} else if ( (strlen(toacc) <= 2) && (toacc[i] == ' ') ) {
	    toret[toreti++] = ' '; toret[toreti++] = 's'; toret[toreti++] = 'p'; 
	    toret[toreti++] = 'a'; toret[toreti++] = 'z'; toret[toreti++] = 'i'; 
	    toret[toreti++] = 'o'; toret[toreti++] = ' ';
	    i++;

	}  else
		toret[toreti++] = toacc[i++];
    }
    
    toret[toreti] = '\0';
    
    strcpy(toacc, toret);
    free(toret);
    
return toacc;
}
	






void load_modules(char *filename) {

    int count=-1, i=0, hcount = -1;
    char *error;
    struct domain *tmpdom = (struct domain *) malloc(sizeof(struct domain));
    struct cell *tmpcell;


    for(i=0; i<maxModules; i++) { 
	mod_fun[count] = NULL;
	strcpy (mod_names[i], "");
	handle[i] = NULL;
    }
    tmpdom = create_rules(filename);

    while ((strcmp(tmpdom->header, "Modules") != 0) && (tmpdom->next != NULL)) 
	 tmpdom = tmpdom->next;

    if ((strcmp(tmpdom->header, "Modules") != 0) && (tmpdom->next == NULL)) return;
    
    tmpcell = tmpdom->headptr->table;
    
    do {
	count++;	hcount++;

        handle[hcount] = dlopen(tmpcell->event, RTLD_NOW);
	if (!handle[hcount]) { printf("dlopen error: %s\n", dlerror()); exit(-1);  }
	
	mod_fun[count] = dlsym(handle[hcount], tmpcell->actions->action);
	if ( (error=dlerror()) != NULL) { printf("dlsym error: %s on %s\n", error, tmpcell->actions->action ); exit(-1); } 
	strcpy(mod_names[count], tmpcell->actions->action);
	
	printf("mod_fun[%d]: = %s()\n", count, mod_names[count]);
    } while ((tmpcell=tmpcell->next_cell) != NULL);

    free(tmpdom);	free(tmpcell);

return;
}


    






int find_pid (int startpid, char *command, char *dev) {
    int i, count, fd, k=1;
    char *retcommand = (char *) malloc(128), *retdev = (char *) malloc(128), buf[128];

    
    for (i=startpid; i<10000; i++, k=1) {    

    	bzero(retcommand, 128);
	sprintf(buf, "/proc/%d/stat", i);
        if ( (fd = open(buf, O_RDONLY)) < 1) continue;    
//        read(fd, retcommand, strlen(command));    
        read(fd, &retcommand[0], 1);
	while (retcommand[0] != '(') read(fd, &retcommand[0], 1);
	read(fd, &retcommand[0], 1);
	while (retcommand[k-1] != ')')  read(fd, &retcommand[k++], 1);
	retcommand[k-1] = '\0';
        close(fd);

        bzero(buf, 128);      sprintf(buf, "/proc/%d/fd/%d", i, 2);
	if((count=readlink(buf, retdev, 1024)) == -1) continue;
        retdev[count] = '\0';

	if ( (strcmp(dev, retdev) == 0) && (strcmp(command, retcommand) == 0) ) {
//    	    printf("PID = %d, command = %s, ret = %s, dev = %s, ret = %s\n\n", i, command, retcommand, dev, retdev);
	    return i;
	}
    }

return -1;
}





int check_EXT(int cons, int x, int y, char token) {	//  molto lenta...

    int fd, i=0, ret = 0;
    char *conname = (char *) malloc(20), *tok = (char *) malloc(2);
    
    sprintf(conname, "/dev/vcs%d", cons);
    if ( (fd = open(conname, O_RDONLY, 0)) < 0 ) { perror("Can't open /dev/vcs"); return 0;}
    lseek(fd, ((y-1)*80)+(x-1), 0);	
    tok[0] = ' ';	read(fd, tok, 1);	tok[1] = '\0';

//    printf("(%d, %d) = %c\n\n", x, y, tok[0]);
    if (token == tok[0]) ret = 1;

    free(conname);	free(tok);
    close(fd);

return ret;
}




int readscreen(char *keycode) {

    char    *tok = (char *) malloc(2), 
	    *line = (char *) malloc(81), 
	    *word = (char *) malloc(80*25);
    int pos = 0, i=0, lsret=0, templine = 0, offs=0;

    char toRead[1024]="";



//    printf("Entering readscreen(%s)\n", keycode);

    if (strcmp(keycode, "1") == 0) {	// read prev char

//	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);	
    
	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);
    
	if ( (vcsOffs[cons]/numcols)*numcols == vcsOffs[cons]) { // ctrl se inizio linea
	    printf("[*1*] Sei gia` al primo carattere\n");
	    speak("Sei gia` al primo carattere\n");
	    free(tok); free(line); free(word); 
	    return -1;
	}
	vcsOffs[cons] = lseek(vcsSR[cons], -1, SEEK_CUR);	    
        tok[0] = ' '; vcsOffs[cons] += read(vcsSR[cons], tok, 1); tok[1] = '\0';
        printf("[*1*] %c\n", tok[0]);
	sprintf(toRead, "%c\n", tok[0]);
	speak(makeAcc(toRead));
	vcsOffs[cons] = lseek(vcsSR[cons], -1, SEEK_CUR);	

        free(tok); free(line); free(word); 
	
	return 0;
    }


    if (strcmp(keycode, "2") == 0) {	// read curr char

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);
	
    	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

        tok[0] = ' '; vcsOffs[cons] += read(vcsSR[cons], tok, 1); tok[1] = '\0';
        printf("[*2*] %c\n", tok[0]);
	sprintf(toRead, "%c\n", tok[0]);
	speak(makeAcc(toRead));
	
        vcsOffs[cons] = lseek(vcsSR[cons], -1, SEEK_CUR);	
	
        free(tok);  free(line); free(word);
	
	return 0;
    }

    if (strcmp(keycode, "3") == 0) {	//read next char

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);

	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

	if ( ((vcsOffs[cons]+1)/numcols)*numcols == (vcsOffs[cons]+1)) { // ctrl se fine linea
	    printf("[*3*] Sei gia` all'ultimo carattere\n");
	    speak("Sei gia` all'ultimo carattere\n");
	    free(tok); free(line); free(word); 
	    return -1;
	}

	vcsOffs[cons] = lseek(vcsSR[cons], +1, SEEK_CUR);
        tok[0] = ' '; vcsOffs[cons] += read(vcsSR[cons], tok, 1); tok[1] = '\0';
        printf("[*3*] %c\n", tok[0]);
	sprintf(toRead, "%c\n", tok[0]);
	speak(makeAcc(toRead));
	vcsOffs[cons] = lseek(vcsSR[cons], -1, SEEK_CUR);

        free(tok);  free(line); free(word);
	
	return 0;
    }

    if (strcmp(keycode, "4") == 0) {	// read prev word

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);

 	lsret = vcsOffs[cons];   lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);
	
	if ( (lsret/numcols)*numcols == lsret) { // ctrl se inizio linea
//	    printf("--1--\n");
	    printf("[*4*] Sei gia` alla prima parola\n");
	    speak("Sei gia` alla prima parola\n");
	    free(tok);
	    return;
	}


	tok[0] = 'X'; 	// torniamo indietro, fino all'inzio della parola
//	printf("((lsret-1)/80)*80 = %d\n(lsret-1) = %d\n", ((lsret-1)/80)*80, (lsret-1) );
	while ( (tok[0] != ' ') && (((lsret)/numcols)*numcols != (lsret)) && (lsret!=-1) ) {
	    tok[0] = ' ';  read(vcsSR[cons], tok, 1);	
	    lsret = lseek(vcsSR[cons], -2, SEEK_CUR);
//	    printf("[1] ret: %d, tok: %c\n", lsret, tok[0]);
	}
//	printf("[inizio word] lsret: %d, tok: %c\n", lsret, tok[0]);
	if (tok[0] == ' ') {		// siamo arrivati agli spazi
//	    printf("caso 1\n");
    	    lsret = lseek(vcsSR[cons], +2, SEEK_CUR);
	} else if (lsret == -1) { 	// inizio schermo
//	    printf("caso 2\n");	
	    lsret = 0;
	} else {			// inizio riga
//	    printf("caso 3\n");	
//	    printf("--2--\n");
	    printf("[*4*] Sei gia` alla prima parola\n");
	    speak("Sei gia` alla prima parola\n");	    
	    free(tok);
	    return;
	}

	    // leviamo gli spazi
	tok[0] = ' '; 	lsret = lseek(vcsSR[cons], -1, SEEK_CUR);
//	printf("((lsret-1)/80)*80 = %d\n(lsret-1) = %d\n", ((lsret-1)/80)*80, (lsret-1) );
	while ( (tok[0] == ' ') && (((lsret)/numcols)*numcols != (lsret)) && (lsret!=-1) ) {
	    tok[0] = 'X';  read(vcsSR[cons], tok, 1);	
	    lsret = lseek(vcsSR[cons], -2, SEEK_CUR);
//	    printf("[1] ret: %d, tok: %c\n", lsret, tok[0]);
	}
//        printf("[leva blank] lsret: %d, tok: %c\n", lsret, tok[0]);
	if (tok[0] != ' ') {		// siamo arrivati alla parola prec
//	    printf("caso 1\n");
    	    lsret = lseek(vcsSR[cons], +2, SEEK_CUR);
	} else if (lsret == -1) { 	// inizio schermo
//	    printf("caso 2\n");	
	    lsret = 0;
	} else {			// inizio riga
//	    printf("caso 3\n");		
//	    printf("--3--\n");
	    printf("[*4*] Sei gia` alla prima parola\n");
	    speak("Sei gia` alla prima parola\n");
	    free(tok);
	    return;
	}

	// torniamo indietro, fino all'inzio della parola
	tok[0] = 'X'; 	lsret = lseek(vcsSR[cons], -2, SEEK_CUR);
//	printf("((lsret-1)/80)*80 = %d\n(lsret-1) = %d\n", ((lsret-1)/80)*80, (lsret-1) );
	while ( (tok[0] != ' ') && (((lsret)/numcols)*numcols != (lsret)) && (lsret!=-1) ) {
	    tok[0] = ' ';  read(vcsSR[cons], tok, 1);	
	    lsret = lseek(vcsSR[cons], -2, SEEK_CUR);
//	    printf("[1] ret: %d, tok: %c\n", lsret, tok[0]);
	}
//	printf("[inizio word] lsret: %d, tok: %c\n", lsret, tok[0]);
	if (tok[0] == ' ') {		// siamo arrivati agli spazi
//	    printf("caso 1\n");
    	    lsret = lseek(vcsSR[cons], +2, SEEK_CUR);
	} else if (lsret == -1) { 	// inizio schermo
//	    printf("caso 2\n");	
	    lsret = 0;
	} else {			// inizio riga
//	    printf("caso 3\n");		
	}

	    // legge la parola
	vcsOffs[cons] = lsret;

	tok[0] = 'X';		i=0;
//	printf("((lsret+1)/80)*80 = %d\n(lsret+1) = %d\n", ((lsret+1)/80)*80, (lsret+1) );
	while ( (tok[0] != ' ') && (((lsret+1)/numcols)*numcols != (lsret+1)) && (lsret!=-1) ) {
	    tok[0] = ' ';	lsret=read(vcsSR[cons], tok, 1);
    	    word[i++] = tok[0];
//	    printf("[2] ret: %d, tok[%d]: %c\n", lsret, i, tok[0]);	    	    
	}
//        printf("[fine word] lsret: %d, tok[%d]: %c\n", lsret, i, tok[0]);	    	    
	word[i] = '\0';

	printf("[*4*] %s\n", word);
	sprintf(toRead, "%s\n", word);
	speak(makeAcc(toRead));
	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

    	
        free(tok);
        free(word);	
    }



    if (strcmp(keycode, "5") == 0) {	// read curr word

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);

 	tok[0] = 'X'; 
	lsret = vcsOffs[cons];	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);
	if ((lsret/numcols)*numcols == lsret) goto salto1;

//	printf("((lsret-1)/80)*80 = %d\n(lsret-1) = %d\n", ((lsret-1)/80)*80, (lsret-1) );
	
	while ( (tok[0] != ' ') && (((lsret)/numcols)*numcols != (lsret)) && (lsret!=-1) ) {
	    tok[0] = ' ';  read(vcsSR[cons], tok, 1);	
	    lsret = lseek(vcsSR[cons], -2, SEEK_CUR);
//	    printf("[1] ret: %d, tok: %c\n", lsret, tok[0]);
	}
	if (tok[0] == ' ') {		// siamo arrivati agli spazi
//	    printf("caso 1\n");
    	    lsret = lseek(vcsSR[cons], +2, SEEK_CUR);
	} else if (lsret == -1) { 	// inizio schermo
//	    printf("caso 2\n");	
	    lsret = 0;
	} else {			// inizio riga
//	    printf("caso 3\n");		
//	    lsret = lseek(vcsSR[cons], +1, SEEK_CUR);
	}
	    
	vcsOffs[cons] = lsret;
	salto1:	


	tok[0] = 'X';		i=0;
	while ( (tok[0] != ' ') && (((lsret+1)/numcols)*numcols != (lsret+1)) && (lsret!=-1) ) {
//	printf("((lsret+1)/80)*80 = %d\n(lsret+1) = %d\n", ((lsret+1)/80)*80, (lsret+1) );
	    tok[0] = ' ';	lsret=read(vcsSR[cons], tok, 1);
    	    word[i++] = tok[0];
//	    printf("[2] ret: %d, tok[%d]: %c\n", lsret, i, tok[0]);	    	    
	}
	word[i] = '\0';

	printf("[*5*] %s\n", word);
	sprintf(toRead, "%s\n", word);
	speak(makeAcc(toRead));

	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);
		
        free(tok);
        free(word);	
    }


    if (strcmp(keycode, "6") == 0) {	// read next word

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);
	
 	tok[0] = 'X'; 
	lsret = vcsOffs[cons];	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

	if (((lsret+1)/numcols)*numcols == (lsret+1)) {
//	    printf("--1--\n");
	    printf("[*6*] Sei gia` all'ultima parola\n");
	    speak("Sei gia` all'ultima parola\n");
	    free(tok);
	    return;
	}

	// va fino alla fine della parola
	while ( (tok[0] != ' ') && (((lsret+1)/numcols)*numcols != (lsret+1)) && (lsret!=-1) ) {
	    tok[0] = ' ';  lsret += read(vcsSR[cons], tok, 1);	
//	    printf("[1] ret: %d, tok: %c\n", lsret, tok[0]);
	}
//	printf("[fine word] lsret: %d, tok: %c\n", lsret, tok[0]);
	if (tok[0] == ' ') {		// siamo arrivati agli spazi
//	    printf("caso 1\n");
//    	    lsret = lseek(vcsSR[cons], +2, SEEK_CUR);
	} else if (lsret == -1) { 	// fine schermo
//	    printf("caso 2\n");	
	    printf("[*6*] Sei gia` all'ultima parola\n");
	    speak("Sei gia` all'ultima parola\n");	    
	    free(tok);
	} else {			// fine riga
//	    printf("caso 3\n");		
//	    printf("--2--\n");
	    printf("[*6*] Sei gia` all'ultima parola\n");
	    speak("Sei gia` all'ultima parola\n");
	    free(tok);
	    return;

	}
	    
	    // leviamo gli spazi
	tok[0] = ' '; 	
//	printf("((lsret-1)/80)*80 = %d\n(lsret-1) = %d\n", ((lsret-1)/80)*80, (lsret-1) );
	while ( (tok[0] == ' ') && (((lsret+1)/numcols)*numcols != (lsret+1)) && (lsret!=-1) ) {
	    tok[0] = 'X';  lsret+=read(vcsSR[cons], tok, 1);	
//	    printf("[1] ret: %d, tok: %c\n", lsret, tok[0]);
	}
//        printf("[leva blank] lsret: %d, tok: %c\n", lsret, tok[0]);
	if (tok[0] != ' ') {		// siamo arrivati alla parola prec
//	    printf("caso 1\n");
    	    lsret = lseek(vcsSR[cons], -1, SEEK_CUR);
	} else if (lsret == -1) { 	// inizio schermo
//	    printf("caso 2\n");	
	    printf("[*6*] Sei gia` all'ultima parola\n");
	    speak("Sei gia` all'ultima parola\n");
	    free(tok);
	} else {			// inizio riga
//	    printf("caso 3\n");		
//	    printf("--3--\n");
	    printf("[*6*] Sei gia` all'ultima parola\n");
	    speak("Sei gia` all'ultima parola\n");
	    free(tok);
	    return;
	}
	
		
	vcsOffs[cons] = lsret;
	
	// legge la parola
	tok[0] = 'X';	i=0;
	while ( (tok[0] != ' ') && (((lsret+1)/numcols)*numcols != (lsret+1)) && (lsret!=-1) ) {
//	printf("((lsret+1)/80)*80 = %d\n(lsret+1) = %d\n", ((lsret+1)/80)*80, (lsret+1) );
	    tok[0] = ' ';	lsret=read(vcsSR[cons], tok, 1);
    	    word[i++] = tok[0];
//	    printf("[2] ret: %d, tok[%d]: %c\n", lsret, i, tok[0]);	    	    
	}
//	printf("[fine word] ret: %d, tok[%d]: %c\n", lsret, i, tok[0]);	    	    	
	word[i] = '\0';

	printf("[*6*] %s\n", word);
	sprintf(toRead, "%s\n", word);
	speak(makeAcc(toRead));

	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);


        free(tok);
        free(word);	
    }



    if (strcmp(keycode, "7") == 0) {	// read prev line

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);
	
	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

	if (vcsOffs[cons] < numcols) {
	    printf("[*7*] Sei gia` alla prima linea\n");
	    speak("Sei gia` alla prima linea\n");
	    free(tok);
	    return;
	}

	pos = vcsOffs[cons];
	lseek(vcsSR[cons], ((pos/numcols)*numcols)-numcols, SEEK_SET);
        read(vcsSR[cons], line, numcols);

        printf("[*7*] %s\n", line);
	sprintf(toRead, "%s\n", line);
	speak(makeAcc(toRead));

	vcsOffs[cons] = ((pos/numcols)*numcols)-numcols;
	lseek(vcsSR[cons], ((pos/numcols)*numcols)-numcols, SEEK_SET);

        free(line);
    }


    if (strcmp(keycode, "8") == 0) {	// read curr line

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);

	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

	pos = vcsOffs[cons];
	lseek(vcsSR[cons], (pos/numcols)*numcols, SEEK_SET);
        read(vcsSR[cons], line, numcols);

        printf("[*8*] %s\n", line);
	sprintf(toRead, "%s\n", line);
	speak(makeAcc(toRead));

	lseek(vcsSR[cons], pos, SEEK_SET);

        free(line);
    }

    if (strcmp(keycode, "9") == 0) {	// read next line

// 	printf("vcsOffs[cons] = %d\n", vcsOffs[cons]);

	lseek(vcsSR[cons], vcsOffs[cons], SEEK_SET);

	if (vcsOffs[cons] >= ((numrows*numcols)-numcols)) {
	    printf("[*9*] Sei gia` all'ultima linea\n");
	    speak("Sei gia` all'ultima linea\n");	    
	    free(tok);
	    return;
	}

	pos = vcsOffs[cons];
	lseek(vcsSR[cons], ((pos/numcols)*numcols)+numcols, SEEK_SET);
        read(vcsSR[cons], line, numcols);

        printf("[*9*] %s\n", line);
	sprintf(toRead, "%s\n", line);
	speak(makeAcc(toRead));

	vcsOffs[cons] = ((pos/numcols)*numcols)+numcols;
	lseek(vcsSR[cons], ((pos/numcols)*numcols)+numcols, SEEK_SET);

        free(line);
    }




return;
}



int speak(char *frase) {

    strcat(festival_buf, frase);
    strcat(festival_buf, "\n");

	if (festivalCommand(info, "(audio_mode 'shutup)\n") != 0) printf("Errore nella Speech!\n");    		    	
	if (festivalCommand(info, userVoice) != 0) printf("Errore nella Speech!\n");    		    	
//    printf("speak *4*\n");
	if (festivalCommand(info, "(Parameter.set \"Duration_Stretch\" 1.2)\n") != 0) printf("Errore nella Speech!\n");    		    	
//    printf("speak *5*\n");
	if (festivalSpeech(info, festival_buf) != 0) printf("Errore nella Speech!\n");    		    
//    printf("speak *6*\n");    
	bzero(festival_buf, 4096);

//    fire_speak(SIGALRM);
    
return 0;
}



static void fire_speak(int signo) {
    
    int tmppid = 0, festpid = 0;

    if (signo==SIGALRM) { 

	festivalClose(info);	// chiude festival e poi lo killa per sicurezza
    	tmppid=find_pid(0, "festival", "/dev/tty12");
	festpid=find_pid(tmppid+1, "festival", "/dev/tty12");
	printf("find_pid(0, \"festival\",\"/dev/tty12\") = %d\n", tmppid);
        printf("find_pid(tmppid+1, \"festival\",\"/dev/tty12\") = %d\n", festpid);    
	if ( festpid > 0 ) kill (festpid, 9);


        info = festival_default_info();
        printf("server: %s, on port: %d, with mode: %s\n", info->server_host, info->server_port, info->text_mode);
//    if ( (info = festivalOpen(info)) == NULL) { printf("Festival Daemon not loaded\n"); exit(-1); }
        if ( (info = festivalOpen(info)) == NULL)  printf("Festival Daemon not loaded\n"); 
	
 
        printf("festival_buf: %s\n", festival_buf);

	if (festivalCommand(info, userVoice) != 0) printf("Errore nella Speech!\n");    		    	
	if (festivalCommand(info, "(Parameter.set \"Duration_Stretch\" 1.2)\n") != 0) printf("Errore nella Speech!\n");    		    	
	if (festivalSpeech(info, festival_buf) != 0) printf("Errore nella Speech!\n");    		    
    
	bzero(festival_buf, 4096);
	    
//        alarm(fest_delay);
    }

return;
}




